@@ -3,6 +3,7 @@ mod key;
 mod dump_stack;
 mod logging;
 mod wrappers;
+mod device_scan;
 
 use std::{
     ffi::{c_char, CString},
@@ -47,6 +48,7 @@ fn handle_c_command(mut argv: Vec<String>, symlink_cmd: Option<&str>) -> i32 {
             "data" => c::data_cmds(argc, argv),
             "device" => c::device_cmds(argc, argv),
             "dump" => c::cmd_dump(argc, argv),
+            "undump" => c::cmd_undump(argc, argv),
             "format" => c::cmd_format(argc, argv),
             "fs" => c::fs_cmds(argc, argv),
             "fsck" => c::cmd_fsck(argc, argv),
